#include "gba.h"
#include "screenmode.h"
#include "timers.h"
#include "dma.h"
#include "bios_div.h"


#include "modtypes.h"
#include "sxmp.h"
#include "sxmp_private.h"

extern const mod_t * const song_table[];

#define DS_A_RATIO_50    0
#define DS_A_RATIO_100   (1<<0x2) 
#define DS_B_RATIO_50    0
#define DS_B_RATIO_100   (1<<0x3) 
        
#define DS_A_RIGHT       (1<<0x8)     
#define DS_A_LEFT        (1<<0x9)
#define DS_A_TIMER0      0
#define DS_A_TIMER1      (1<<0xa)
#define DS_A_FIFO        (1<<0xb)

#define DS_B_RIGHT       (1<<0xc)
#define DS_B_LEFT        (1<<0xd)        
#define DS_B_TIMER0      0
#define DS_B_TIMER1      (1<<0xe)
#define DS_B_FIFO        (1<<0xf)

sound_var_t g_sound_vars = {};
sound_channel_t g_channel[CHANNEL_COUNT] = { };

s8 g_snd_buffer[2*BUFFER_SIZE] IN_EWRAM;
s8 g_snd_buffer_2[2*BUFFER_SIZE] IN_EWRAM;
/*
s8 g_snd_buffer[2*BUFFER_SIZE] ;
s8 g_snd_buffer_2[2*BUFFER_SIZE] ;
*/

typedef struct
{
  const u8 * order; // pointer to ROM table of orders
 
  const pattern_t * pattern; // pointer to current pattern array 
  const u8 * pattern_data;   // shortcut to the data


  u8 current_pattern;
  u8 max_order;   // where the song ends
  u8 order_index; // where we are in the xm order
  u8 ticks;        // ticking

  u8 current_row;  // when this reaches pattern.row, then go to the next row
  u8 speed;        // current speed - ticks per row
  u8 tempo;        // current tempo in BPM
  u8 state;        // playing paused whatever
  
  u8 song;         // song id
  u8 channels;         // song id
  u8 loop;         // loop when finished
  
  u8 fade;         // if 0, then no fade. if > 0 then fade by this amount
} gb_mod_t;

gb_mod_t g_mod_track;

extern const unsigned int note_table[];

typedef struct
{
  int channel;
  int note;
  int relative;
  int finetune;
  int volume;
  int effect;
  int param;
  
}note_t;

static void xm_note(note_t * xmnote)
{
  if (xmnote->note) {
    int channel = xmnote->channel;

    if (xmnote->note != 97) {
      if (channel >= CHANNEL_COUNT) {
        // channel isn't available. look for a free one
        // OR check if this channel has already hijacked one
        //int chin = channel;
        int i;
        for (i = CHANNEL_COUNT-1; i >= 0; i--) {
          if(g_channel[i].channel==channel) {
            channel = i;
            xmnote->channel = i;
            break;
          }
        }
        if (i < 0) {
          for (i = CHANNEL_COUNT-1; i >=0; i--) {
            if(g_channel[i].data==0) {
              g_channel[i].channel = channel;
              channel = i;
              xmnote->channel = i;
              break;
            }
          }
        }
        if (i == -1)
          goto effects;
      } else {
        g_channel[channel].channel = channel;
      }

      int adjusted_note = xmnote->note-1+xmnote->relative;
      // finetune between notes
      int a = note_table[adjusted_note];
      int b = note_table[adjusted_note+1];
      adjusted_note = (xmnote->finetune*(b-a)/256)+a;

      g_channel[channel].increment = adjusted_note; // note_table[adjusted_note];
      g_channel[channel].position = 0;
      if (g_mod_track.fade) {
        g_channel[channel].volume = (xmnote->volume*g_mod_track.fade)/64;
      } else {
        g_channel[channel].volume = xmnote->volume;
      }
    } else {
      if (channel >= CHANNEL_COUNT) {
        // note off - make sure we catch hijack channels!
        //int chin = channel;
        int i;
        for (i = CHANNEL_COUNT-1; i >= 0; i--) {
          if(g_channel[i].channel==channel) {
            channel = i;
            xmnote->channel = i;
            break;
          }
        }
        if (i == -1)
          goto effects;

      }
      g_channel[channel].loop_length = 0;
    }
  }
effects:
  if (xmnote->effect) {
    if (xmnote->effect == 0xf) {
      if (xmnote->param > 0x20) {
        g_mod_track.tempo = xmnote->param;
        g_sound_vars.samples_per_clock = bios_divide(SOUND_FREQUENCY,(g_mod_track.tempo*2/5)); 
      } else {
        g_mod_track.speed = xmnote->param;
      }
    } 
  }
}

static void xm_row() 
{
  // new row
  int ch;
  for (ch = 0; ch < g_mod_track.channels; ch++) {
    u8 note =0, sample=0, volume=0, effect=0, param=0;
    // unpack pattern data - uses XM/s3m mix
    u8 what = *g_mod_track.pattern_data++;
    if (what) {
      if (what&0x80) {
        if (what&1)
          note   = *g_mod_track.pattern_data++;
        if (what&2)
          sample = *g_mod_track.pattern_data++;
        if (what&4)
          volume = *g_mod_track.pattern_data++;
        if (what&8) 
        {
          effect = *g_mod_track.pattern_data++;
          param  = *g_mod_track.pattern_data++;
        }
      }
      else {
        note   = what;
        sample = *g_mod_track.pattern_data++;
        volume = *g_mod_track.pattern_data++;
        effect = *g_mod_track.pattern_data++;
        param  = *g_mod_track.pattern_data++;
      }
    } else {
      break;
    }
#if 0
    if (note|effect|sample|volume) {
      gbaprintf("%d,%d,%d,%d,%d,%d\n",
          note, sample, volume, effect,param,ch );
    }
#endif
    // if all are 0 then do nothing
    if (note || effect) {
      note_t n = {
          ch,
          note,
          song_table[g_mod_track.song]->sample[sample-1].relative,
          song_table[g_mod_track.song]->sample[sample-1].finetune,
          volume,
          effect,
          param
      };
      xm_note(&n);
      if (sample && n.channel >= 0 && n.channel < CHANNEL_COUNT) {
        // fetch sample from list
        g_channel[n.channel].data = song_table[g_mod_track.song]->sample[sample-1].data;
        g_channel[n.channel].length = song_table[g_mod_track.song]->sample[sample-1].length<<12;
        g_channel[n.channel].loop_length = song_table[g_mod_track.song]->sample[sample-1].loop_length<<12;
      }
    }
  }
}

static void xm_update()
{
  if (g_mod_track.state != SXMP_PLAYING)
    return;

  if (g_mod_track.fade) {
    g_mod_track.fade--;
    if (!g_mod_track.fade) {
      sxmp_stop_song();
    }
  }
        
  // tick, if tick < speed increment row
  if (++g_mod_track.ticks >= g_mod_track.speed) {

    g_mod_track.ticks = 0;
    if (++g_mod_track.current_row >= g_mod_track.pattern[g_mod_track.current_pattern].rows) {
      g_mod_track.current_row = 0;
      if (++g_mod_track.order_index >= g_mod_track.max_order) {
        g_mod_track.state = SXMP_STOPPED;
        if (g_mod_track.loop) {
          sxmp_start_song(g_mod_track.song, g_mod_track.loop);
        }
        return;
      }
      else {
        g_mod_track.current_pattern = g_mod_track.order[g_mod_track.order_index];
        g_mod_track.pattern_data = g_mod_track.pattern[g_mod_track.current_pattern].data;
      }
    }

    xm_row();
    
  }
}

void sxmp_update()
{
  s32 samples_left = g_sound_vars.mix_buffer_length;  // bytes left in the buffer
  while  (samples_left > 0) {
    if (g_sound_vars.samples_until_clock == 0) 
    {
      // update the midi, set the bytes to mix before next update
#if USE_PROFILING
      BGPaletteMem[0] = RGB16(31,21,0);
#endif
      xm_update();
#if USE_PROFILING
      BGPaletteMem[0] = RGB16(0,0,0);
#endif
      g_sound_vars.samples_until_clock =  g_sound_vars.samples_per_clock;
    }
    // is this the last batch of samples for this frame?
    if (g_sound_vars.samples_until_clock < samples_left) 
    {
      // not enough samples to fill the buffer, will need to update position in midi
      // mix the samples required until the next midi event
      if (g_mod_track.state != SXMP_PLAYING)
        sxmp_no_channel_mix(g_sound_vars.samples_until_clock);
      else
        sxmp_sound_mix(g_sound_vars.samples_until_clock);
      samples_left -= g_sound_vars.samples_until_clock;
      // now we have mixed what we had left, next time through this loop will update midi position
      g_sound_vars.samples_until_clock = 0;
    }
    else
    {
      // enough samples left to fill the buffer but not to update midi, mix what we have
      if (g_mod_track.state != SXMP_PLAYING)
        sxmp_no_channel_mix(samples_left);
      else
        sxmp_sound_mix(samples_left);
      g_sound_vars.samples_until_clock -= samples_left;
      samples_left = 0;
    }
  }
}

void sxmp_start_song(int song, int loop) 
{
  g_mod_track.loop = loop;
  g_mod_track.song = song;
  g_mod_track.order = song_table[song]->order;
  g_mod_track.pattern = song_table[song]->pattern;
  g_mod_track.order_index = 0;
  g_mod_track.current_row = 0;
  g_mod_track.speed = song_table[song]->speed;
  g_mod_track.tempo = song_table[song]->tempo;
  g_mod_track.state = SXMP_PLAYING;
  g_mod_track.current_pattern = g_mod_track.order[g_mod_track.order_index];
  g_mod_track.pattern_data = song_table[song]->pattern[g_mod_track.current_pattern].data;
  g_mod_track.ticks = 0;
  g_mod_track.max_order = song_table[song]->song_length;
  g_sound_vars.samples_per_clock = SOUND_FREQUENCY/(g_mod_track.tempo*2/5); 
  g_mod_track.channels = song_table[song]->channels;
  g_mod_track.fade = 0;

  int i;
  for (i = 0; i < CHANNEL_COUNT; i++) {
    g_channel[i].channel = i;
    g_channel[i].data = 0;
    g_channel[i].position = 0;
    g_channel[i].increment = 0;
  }
}

int sxmp_is_playing() 
{
  return (g_mod_track.state == SXMP_PLAYING);
}

void sxmp_stop_song() 
{
  g_mod_track.state = SXMP_STOPPED;
  sxmp_sound_stop();
}
void sxmp_fade_out() 
{
  g_mod_track.fade = 32;
}

void sxmp_fake_stereo(int fake)
{
  if (fake) {
    REG_SGCNT0_H =  DS_A_RATIO_100 |  /*DS_A_RIGHT |*/ DS_A_LEFT | DS_A_FIFO
      | DS_B_RATIO_100 |  DS_B_RIGHT | /*DS_B_LEFT |*/ DS_B_FIFO ;
  }
  else {
    REG_SGCNT0_H =  DS_A_RATIO_100 |  DS_A_RIGHT | DS_A_LEFT | DS_A_FIFO
      | DS_B_RATIO_100 |  DS_B_RIGHT | DS_B_LEFT | DS_B_FIFO ;
  }
    
}

void sxmp_init() 
{
  // put the magic numbers in regs
  sxmp_fake_stereo(1);
  REG_SGCNT1   = 0x80;

  // empty the buffer...
  int i = 0;
  REG_DMA3SAD = (u32)(&i);
  REG_DMA3DAD = (u32)(g_snd_buffer);
  REG_DMA3CNT_L = (2*BUFFER_SIZE/4);  // 4 bytes per word
  REG_DMA3CNT_H = ((DMA_SOURCE_FIXED|DMA_32NOW)>>16);
  REG_DMA3SAD = (u32)(&i);
  REG_DMA3DAD = (u32)(g_snd_buffer_2);
  REG_DMA3CNT_L = (2*BUFFER_SIZE/4);  // 4 bytes per word
  REG_DMA3CNT_H = ((DMA_SOURCE_FIXED|DMA_32NOW)>>16);

  // set mixer variables
  g_sound_vars.mix_buffer = g_snd_buffer;
  g_sound_vars.mix_buffer_2 = g_snd_buffer_2;
  g_sound_vars.mix_buffer_length = BUFFER_SIZE;
  g_sound_vars.active = 1;
  g_sound_vars.current_mix_buffer = g_sound_vars.mix_buffer;

  // initial value = default
  g_sound_vars.samples_per_clock = SOUND_FREQUENCY/(120*2/5); 
  g_mod_track.state = SXMP_STOPPED;
  g_mod_track.fade = 0;

  sxmp_sound_stop();

  // start timer 0 (why?)
  REG_TM0D = (65536 - (16777216 / SOUND_FREQUENCY));
  REG_TM0CNT = TIME_ENABLE;
  //SetTimerControl(0,TIME_ENABLE);
  // initialise DMA1 destination
  REG_DMA1CNT_H = 0;
  REG_DMA1DAD = (u32) &REG_SGFIFOA; //0x40000A0;
  
  REG_DMA2CNT_H = 0;
  REG_DMA2DAD = (u32) &REG_SGFIFOB; //0x40000A4;

}

void sxmp_vsync() 
{
#if 1
  if (g_sound_vars.active == 1) 
  {
    // start buffer 0
    REG_DMA1CNT_H = 0;
    REG_DMA1SAD = (u32)g_sound_vars.mix_buffer;
    REG_DMA1CNT_H = 0xB640;//DMA_DEST_FIXED | DMA_REPEAT | DMA_WORD | DMA_MODE_FIFO | DMA_ENABLE;
    
    REG_DMA2CNT_H = 0;
    REG_DMA2SAD = (u32)g_sound_vars.mix_buffer_2;
    REG_DMA2CNT_H = 0xB640;//DMA_DEST_FIXED | DMA_REPEAT | DMA_WORD | DMA_MODE_FIFO | DMA_ENABLE;

    // make sure we write to the "background"
    g_sound_vars.current_mix_buffer = g_sound_vars.mix_buffer + g_sound_vars.mix_buffer_length;
    g_sound_vars.current_mix_buffer_2 = g_sound_vars.mix_buffer_2 + g_sound_vars.mix_buffer_length;
    g_sound_vars.active = 0;
  }
  else 
  {
    // DMA1 already points to other buffer, so no swap needed
    g_sound_vars.current_mix_buffer = g_sound_vars.mix_buffer;
    g_sound_vars.current_mix_buffer_2 = g_sound_vars.mix_buffer_2;
    g_sound_vars.active = 1;
  }
#endif
}

void sxmp_sound_stop()
{
  int i;
  for (i = 0; i < CHANNEL_COUNT; i++) {
    g_channel[i].data = 0;
    g_channel[i].loop_length = 0;
    g_channel[i].volume = 0;
  }
}

void sxmp_pause(int pause) 
{
  if (pause && g_mod_track.state == SXMP_PLAYING) {
    g_mod_track.state = SXMP_PAUSED;
    // should disable DMA?
    sxmp_sound_stop();
  } else if ( g_mod_track.state == SXMP_PAUSED ) {
    g_mod_track.state = SXMP_PLAYING;
  }
}

int sxmp_paused()
{
  return (g_mod_track.state == SXMP_PAUSED);
}
